import io
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from shutil import copyfileobj
from typing import Optional, Tuple, IO, Union, Any, Callable

import subprocess  # nosec

# Here bandit warns about:
# B404:blacklist import_subprocess
# B603:subprocess_without_shell_equals_true
# 1. We do not use shell=True in this package
# 2. All calls to subprocess start with gpg, so possibly unsanitized input would
#    only result in a gpg error message

if hasattr(subprocess, "STARTUPINFO"):

    def startupinfo():
        _startupinfo = subprocess.STARTUPINFO()
        _startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        _startupinfo.wShowWindow = subprocess.SW_HIDE
        return _startupinfo


else:

    def startupinfo():
        return None


class GPGError(BaseException):
    pass


def raise_from_future(f: Future):
    e = f.exception()
    if e is not None:
        raise e


def cmd(command: Tuple[str, ...], out: Optional[IO[bytes]] = None, **kwargs):
    if out is not None:
        return cmd_devnull(command, stdout=out, **kwargs)
    return cmd_pipe_stdout(command, **kwargs)


def cmd_devnull(
    command: Tuple[str, ...], src=None, stdout=subprocess.DEVNULL, **kwargs
):
    with cmd_pipe(command, src=src, stdout=stdout, **kwargs):
        pass


@contextmanager
def cmd_pipe_stdout(*args, **kwargs):
    with cmd_pipe(*args, **kwargs) as proc:
        yield proc.stdout


@contextmanager
def cmd_pipe(
    command: Tuple[str, ...],
    src: Optional[Union[bytes, io.FileIO, Callable]] = None,
    stdout=subprocess.PIPE,
    passphrase=None,
    **kwargs,
):
    feed = None
    if src is not None:
        if has_fileno(src) and passphrase is None:
            kwargs["stdin"] = src
        else:
            kwargs["stdin"] = subprocess.PIPE
            if isinstance(src, bytes):
                feed = feed_from_str(src)
            elif callable(src):
                feed = feed_from_callable(src)
            else:
                feed = feed_from_stream(src)
    sink = None
    if callable(stdout):
        sink = stdout
        stdout = subprocess.PIPE

    # For the bandit vulnerability suppressed here (nosec):
    # see info close the subprocess import
    with subprocess.Popen(  # nosec
        command,
        stderr=subprocess.PIPE,
        stdout=stdout,
        startupinfo=startupinfo(),
        **kwargs,
    ) as proc:
        if proc.stdin is not None:
            if passphrase is not None:
                proc.stdin.write(passphrase.encode() + b"\n")
                proc.stdin.flush()
            if feed is not None and sink is not None:
                with ThreadPoolExecutor(max_workers=2) as executor:
                    future_feed = executor.submit(feed, proc.stdin)
                    future_sink = executor.submit(sink, proc.stdout)
                    future_feed.add_done_callback(lambda _: proc.stdin.close())  # type: ignore
                    raise_from_future(future_feed)
                    raise_from_future(future_sink)
            elif feed is not None:
                feed(proc.stdin)
                proc.stdin.close()
            elif sink is not None:
                sink(proc.stdout)
        yield proc
        err = proc.stderr.read() if proc.stderr is not None else b""
    if proc.returncode != 0:
        raise GPGError(err.decode("utf-8", "replace"))


def stderr_lookahead(proc: subprocess.Popen) -> bytes:
    if proc.stderr is not None:
        out = proc.stderr.read()
        proc.stderr.close()
        proc.stderr = io.BytesIO(out)
    else:
        out = b""
    return out


class ExpectProc:
    def __init__(self, proc):
        self.proc = proc
        self.source = proc.stderr
        self.dest = proc.stdin
        self.stdout = proc.stdout

    def expect(self, expected, prefix=b"[GNUPG:] GET_"):
        actual = b""
        while not actual.startswith(prefix):
            actual = self.source.read(len(prefix))
            c = True
            while c and c != b"\n":
                c = self.source.read(1)
                if c != b"\r":  # Skip '\r' (windows)
                    actual += c
            if not actual:
                raise ValueError("Unexpected end of source")
        if actual != expected:
            raise ValueError(f"Unexpected prompt from gpg:\n{expected}\n{actual}")

    def put(self, msg):
        self.dest.write(msg)
        self.dest.flush()


@contextmanager
def expect(command, **kwargs):
    # For the bandit vulnerability suppressed here (nosec):
    # see info close the subprocess import
    with subprocess.Popen(  # nosec
        command,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        **kwargs,
    ) as proc:
        yield ExpectProc(proc)


def feed_from_str(src: bytes):
    def feed(stdin):
        stdin.write(src)

    return feed


def feed_from_stream(src: io.FileIO):
    def feed(stdin):
        copyfileobj(src, stdin)

    return feed


def feed_from_callable(src: Callable):
    def feed(stdin):
        src(stdin)

    return feed


def has_fileno(s: Any) -> bool:
    if hasattr(s, "fileno"):
        try:
            s.fileno()
        except BaseException:
            return False
        return True
    return False



from typing import Type, Dict, List, Callable, Generator, Tuple


def compile_grammar(grammar: Dict, final_token: Type) -> Callable:
    """Returns a token parser that transforms an initial sequence of tokens
    by consecutively applying a set of rules to the sequence.
    A rule consists of a pattern of token types and a transforming function.
    When the initial sequence reduces to the final token (identified
    by its "final_token" type), the processing of tokens is stopped.

    :param grammar: dict of the form "pattern": "rule", where each pattern is a
        tuple of tokens that correspond to two consecutive lines of output of
        the gpg command.
    :param final_token: object that has a different type than any other "token"
        in the list of "tokens" passed to the "process" function returned
        by compile_grammar. Typically the final token is a list.
    :return: function that applies rules to a list of tokens.
    """
    grammar = expand_token_sequences(grammar)
    intermediate_token_sequences = frozenset(
        {s[:i] for s in grammar for i in range(1, len(s))} | {()}
    ) - frozenset(grammar)

    def scan_pattern_at_start(tokens: Generator):
        """Look at the beginning of the tokens sequence. If the beginning
        maches a pattern associated to a rule in the grammar, then this rule is
        returned.

        :param tokens: generator of token.
        :return: rule (function) to be applied to the matched token subsequence
        """
        token_pattern: Tuple[type, ...] = ()
        while token_pattern in intermediate_token_sequences:
            try:
                token_pattern += (type(next(tokens)),)
            except StopIteration:
                return None, None
        rule = grammar.get(token_pattern)
        if rule is not None:
            return rule, len(token_pattern)
        return None, None

    def scan_pattern(tokens):
        """General form of :scan_pattern_at_start:. Also matches in the interior
        of the token sequence
        """
        for pos in range(len(tokens)):
            rule, L = scan_pattern_at_start(iter(tokens[pos:]))
            if rule is not None:
                return rule, slice(pos, pos + L)
        return None

    def process(tokens: List):
        """Process tokens in the tokens sequence until there is only one token
        left whose type matches the type of final token.
        """
        while [type(token) for token in tokens] != [final_token]:
            try:
                rule, slc = scan_pattern(tokens)
            except TypeError as e:
                raise ValueError(
                    "Unexpected sequence of tokens:", [type(token) for token in tokens]
                ) from e
            # Apply the rule to the matched subsequence
            subs = rule(*tokens[slc])

            tokens = (
                tokens[: slc.start] + subs + tokens[slc.stop :]
            )  # pylint: disable=invalid-slice-index
        return tokens[0]

    return process


def expand_token_sequences(grammar):
    return {
        key: val for keys, val in grammar.items() for key in expand_token_sequence(keys)
    }


def expand_token_sequence(seq):
    distinction_found = False
    for i, t in enumerate(seq):
        if isinstance(t, tuple):
            distinction_found = True
            for sub_t in t:
                yield from expand_token_sequence(seq[:i] + (sub_t,) + seq[i + 1 :])
    if not distinction_found:
        yield seq



import unittest
import typing
import sys
from dataclasses import dataclass

from gpg_lite import deserialize

# deserialize is shadowed by deserialize.deserialize:
deserialize = sys.modules["gpg_lite.deserialize"]


class TestDeserialize(unittest.TestCase):
    def test_dict_type(self):
        self.assertEqual(deserialize.dict_type(typing.Dict[int, str]), (int, str))
        self.assertEqual(deserialize.dict_type(int), None)

    def test_deserialize(self):
        @dataclass
        class Y:
            x: int
            y: typing.Tuple[bool]

        @dataclass
        class X:
            a: int
            pack: Y

        data = {"a": 1, "pack": {"x": 1, "y": [True, False]}}
        x = deserialize.deserialize(X)(data)

        self.assertEqual(x, X(a=1, pack=Y(x=1, y=(True, False))))

        # typing.Dict
        @dataclass
        class Z:
            a: int
            dct: typing.Dict[int, str]

        data = {"a": 1, "dct": {1: "x", 2: "y"}}
        z = deserialize.deserialize(Z)(data)

        self.assertEqual(z, Z(a=1, dct=data["dct"]))

        # typing.Optional
        self.assertEqual(deserialize.deserialize(typing.Optional[int])(None), None)
        self.assertEqual(deserialize.deserialize(typing.Optional[int])(1), 1)



import unittest
import os
import tempfile
import io
from functools import partial

import gpg_lite as gpg

WITH_GPG = os.environ.get("WITH_GPG", None) == "true"
KEY_SERVER = os.environ.get("KEY_SERVER", None)

CN_UID = gpg.Uid(full_name="Chuck Norris", email="chuck.norris@roundhouse.gov")

UMLAUTS_UID = gpg.Uid(full_name="With Umläuts", email="with.umlauts@umlauts.info")


class TestUtils(unittest.TestCase):
    def test_split_url(self):
        self.assertEqual(
            gpg.split_host("http://keyserver.dcc.sib.swiss:11371"),
            ("http://", "keyserver.dcc.sib.swiss", ":11371"),
        )
        self.assertEqual(
            gpg.split_host("hkp://keyserver.dcc.sib.swiss:11371"),
            ("hkp://", "keyserver.dcc.sib.swiss", ":11371"),
        )
        self.assertEqual(
            gpg.split_host("keyserver.dcc.sib.swiss:11371"),
            (None, "keyserver.dcc.sib.swiss", ":11371"),
        )
        self.assertEqual(
            gpg.split_host("http://keyserver.dcc.sib.swiss"),
            ("http://", "keyserver.dcc.sib.swiss", None),
        )
        self.assertEqual(
            gpg.split_host("keyserver.dcc.sib.swiss"),
            (None, "keyserver.dcc.sib.swiss", None),
        )
        with self.assertRaises(ValueError):
            gpg.split_host("keyserver.dcc:sib.swiss")

    def test_uid_missging_both_fields(self):
        with self.assertRaises(ValueError):
            gpg.Uid(full_name=None, email=None)

    def test_uid_from_str(self):
        self.assertEqual(gpg.Uid.from_str("[User ID not found]"), None)
        self.assertEqual(gpg.Uid.from_str("[User-ID nicht gefunden]"), None)
        self.assertEqual(gpg.Uid.from_str("[Identité introuvable]"), None)
        self.assertEqual(
            gpg.Uid.from_str("Chuck Norris <chucknorris@roundhouse.gov>"),
            gpg.Uid(full_name="Chuck Norris", email="chucknorris@roundhouse.gov"),
        )
        self.assertEqual(
            gpg.Uid.from_str(
                "Chuck Norris (Chucks's test PGP key) " "<chucknorris@roundhouse.gov>"
            ),
            gpg.Uid(
                full_name="Chuck Norris (Chucks's test PGP key)",
                email="chucknorris@roundhouse.gov",
            ),
        )
        self.assertEqual(
            gpg.Uid.from_str("Chuck Norris"),
            gpg.Uid(full_name="Chuck Norris", email=None),
        )
        self.assertEqual(
            gpg.Uid.from_str("chucknorris@roundhouse.gov"),
            gpg.Uid(full_name=None, email="chucknorris@roundhouse.gov"),
        )

    def test_uid_repr(self):
        cases = (
            (
                "Chuck Norris (Chucks's test PGP key)",
                "chucknorris@roundhouse.gov",
                "Chuck Norris (Chucks's test PGP key) <chucknorris@roundhouse.gov>",
            ),
            (
                "Chuck Norris",
                "chucknorris@roundhouse.gov",
                "Chuck Norris <chucknorris@roundhouse.gov>",
            ),
            ("Chuck Norris", None, "Chuck Norris"),
            (None, "chucknorris@roundhouse.gov", "chucknorris@roundhouse.gov"),
        )
        for name, email, expected in cases:
            self.assertEqual(repr(gpg.Uid(full_name=name, email=email)), expected)

    def test_normalize_fingerprint(self):
        for length in (8, 16, 32, 40):
            fp = "A" * length
            self.assertEqual(gpg.normalize_fingerprint(fp), f"0x{fp}")
            self.assertEqual(gpg.normalize_fingerprint(f"0x{fp}"), f"0x{fp}")
        self.assertEqual(gpg.normalize_fingerprint("123 af  47a"), "0x123af47a")
        with self.assertRaises(ValueError):
            gpg.normalize_fingerprint("A" * 7)
        with self.assertRaises(ValueError):
            gpg.normalize_fingerprint("Z" * 8)


class TestGPG(unittest.TestCase):
    def test_factory(self):
        class TestFactory(gpg.Factory):
            type_ = object

        fct = TestFactory()
        fct.append("keys", "A")
        self.assertEqual(fct.kwargs, {"keys": ("A",)})
        fct.append("keys", "B")
        self.assertEqual(fct.kwargs, {"keys": ("A", "B")})

    def test_parse_search_keys(self):
        response = io.BytesIO(
            b"""info:1:1
pub:CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:1:2048:1550241679::
uid:Chuck Norris <chuck.norris@roundhouse.gov>:1550241679::

pub:F119E26211119BB09EFF879D6D229C51AB5D107C:1:2048:1550241679::
uid:With Uml%C3%A4uts <with.umlauts@umlauts.info>:1550241679::
"""
        )
        keys = list(gpg.parse_search_keys(response))
        self.assertEqual(
            keys,
            [
                gpg.KeyInfo(
                    uid=CN_UID,
                    fingerprint="CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107",
                    key_algorithm=1,
                    key_length=2048,
                    creation_date="1550241679",
                    expiration_date=None,
                ),
                gpg.KeyInfo(
                    uid=UMLAUTS_UID,
                    fingerprint="F119E26211119BB09EFF879D6D229C51AB5D107C",
                    key_algorithm=1,
                    key_length=2048,
                    creation_date="1550241679",
                    expiration_date=None,
                ),
            ],
        )

    def test_parse_keys(self):
        key_payload = io.BytesIO(
            b"""tru::1:1571400791:0:3:1:5
pub:u:2048:1:FDC66B7A5F6FFD4B:1550241679:::u:::scESC::::::23:1571230031:1 http\\x3a//keyserver.dcc.sib.swiss\\x3a11371:
fpr:::::::::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:
uid:u::::1550241679::B1E8D4565D26BED47D1EB8BB2B5733AEC1F154B5::Chuck Norris <chuck.norris@roundhouse.gov>::::::::::0:
sig:::1:FDC66B7A5F6FFD4B:1550241679::::Chuck Norris <chuck.norris@roundhouse.gov>:13x::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:::8:
sig:::1:3A6500D5C1DE39AC:1571225229::::DCC Test (Test key for transferprotocol) <dcc@sib.swiss>:13x:::::2:
sig:?::1:2E3E7FE28FE412AE:1573661452::::[User ID not found]:13x::2866A142A16F54E06EB73CF92E3E7FE28FE412AE:::8:
sub:u:2048:1:D892C41917B20115:1550241679::::::e::::::23:
fpr:::::::::55C5314BB9EFD19AE7CC4774D892C41917B20115:
pub:-:4096:1:3A6500D5C1DE39AC:1565257866:::-:::scESC::::::23::0:
fpr:::::::::848FA2707A1AC6060FFE5D253A6500D5C1DE39AC:
uid:-::::1565257866::4D51176EBCDF1F58C3D7E0F76625B168A184BB8E::DCC Test (Test key for transferprotocol) <dcc@sib.swiss>::::::::::0:
sub:-:4096:1:A4C037FA518B5C4F:1565257866::::::e::::::23:
fpr:::::::::3AC9B4462D255449989A3D6EA4C037FA518B5C4F:
"""
        )
        keys = gpg.parse_keys(key_payload)
        dcc_uid = gpg.Uid(
            full_name="DCC Test (Test key for transferprotocol)", email="dcc@sib.swiss"
        )
        self.assertEqual(
            keys,
            [
                gpg.Key(
                    key_id="FDC66B7A5F6FFD4B",
                    fingerprint="CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107",
                    validity=gpg.Validity.ultimately_valid,
                    key_length=2048,
                    pub_key_algorithm=1,
                    creation_date="1550241679",
                    signatures=(
                        gpg.Signature(
                            issuer_uid=CN_UID,
                            issuer_key_id="FDC66B7A5F6FFD4B",
                            creation_date="1550241679",
                            signature_class="13x",
                        ),
                        gpg.Signature(
                            issuer_uid=dcc_uid,
                            issuer_key_id="3A6500D5C1DE39AC",
                            creation_date="1571225229",
                            signature_class="13x",
                        ),
                        gpg.Signature(
                            issuer_uid=None,
                            issuer_key_id="2E3E7FE28FE412AE",
                            creation_date="1573661452",
                            signature_class="13x",
                        ),
                    ),
                    key_capabilities={
                        gpg.KeyCapability.Sign(True),
                        gpg.KeyCapability.Sign(False),
                        gpg.KeyCapability.Certify(True),
                        gpg.KeyCapability.Certify(False),
                        gpg.KeyCapability.Encrypt(True),
                    },
                    key_type=gpg.KeyType.public,
                    uids=(CN_UID,),
                    owner_trust="u",
                    sub_keys=(
                        gpg.SubKey(
                            key_type=gpg.KeyType.public,
                            key_id="D892C41917B20115",
                            fingerprint="55C5314BB9EFD19AE7CC4774D892C41917B20115",
                            validity=gpg.Validity.ultimately_valid,
                            key_length=2048,
                            pub_key_algorithm=1,
                            creation_date="1550241679",
                            key_capabilities={gpg.KeyCapability.Encrypt(False)},
                        ),
                    ),
                    origin="http://keyserver.dcc.sib.swiss:11371",
                ),
                gpg.Key(
                    key_id="3A6500D5C1DE39AC",
                    fingerprint="848FA2707A1AC6060FFE5D253A6500D5C1DE39AC",
                    validity=gpg.Validity.unknown,
                    key_length=4096,
                    pub_key_algorithm=1,
                    creation_date="1565257866",
                    key_capabilities={
                        gpg.KeyCapability.Sign(True),
                        gpg.KeyCapability.Sign(False),
                        gpg.KeyCapability.Certify(True),
                        gpg.KeyCapability.Certify(False),
                        gpg.KeyCapability.Encrypt(True),
                    },
                    key_type=gpg.KeyType.public,
                    uids=(dcc_uid,),
                    owner_trust="-",
                    sub_keys=(
                        gpg.SubKey(
                            key_type=gpg.KeyType.public,
                            key_id="A4C037FA518B5C4F",
                            fingerprint="3AC9B4462D255449989A3D6EA4C037FA518B5C4F",
                            validity=gpg.Validity.unknown,
                            key_length=4096,
                            pub_key_algorithm=1,
                            creation_date="1565257866",
                            key_capabilities={gpg.KeyCapability.Encrypt(False)},
                        ),
                    ),
                    origin=None,
                ),
            ],
        )

    def test_parse_keys_revocation_sig(self):
        key_payload = io.BytesIO(
            b"""
pub:-:2048:1:0000000000000000:1587731664:::-:::scESC::::::23:1588834129:1 http\\x3a//keyserver.dcc.sib.swiss\\x3a11371:
fpr:::::::::1111111111111111111111110000000000000000:
uid:-::::1587731664::6350A25D9F42F55F19A335C67AFC54D0DF542625::Gunnery Sgt. Hartmann <hartmann@bullshit.gov>:::::::::1588834129:1:
sig:::1:0000000000000000:1587731664::::Gunnery Sgt. Hartmann <hartmann@bullshit.gov>:13x:::::2:
sig:::1:881685B5EE0FCBD3:1588064955::::SPHN Data Coordination Centre <dcc@sib.swiss>:13x::B37CE2A101EBFA70941DF885881685B5EE0FCBD3:::8:
rev:::1:881685B5EE0FCBD3:1588066811::::SPHN Data Coordination Centre <dcc@sib.swiss>:30x,00::B37CE2A101EBFA70941DF885881685B5EE0FCBD3:::8:::::revsig\\nInvalid email:
sub:-:2048:1:AAAAAAAAAAAAAAAA:1587731664::::::e::::::23:
fpr:::::::::BBBBBBBBBBBBBBBBBBBBBBBBAAAAAAAAAAAAAAAA:
"""
        )
        keys = gpg.parse_keys(key_payload)
        dcc_uid = gpg.Uid(
            full_name="SPHN Data Coordination Centre", email="dcc@sib.swiss"
        )
        uid = SGT_HARTMANN_UID
        self.assertEqual(
            keys,
            [
                gpg.Key(
                    key_id="0000000000000000",
                    fingerprint="1111111111111111111111110000000000000000",
                    validity=gpg.Validity.unknown,
                    key_length=2048,
                    pub_key_algorithm=1,
                    creation_date="1587731664",
                    signatures=(
                        gpg.Signature(
                            issuer_uid=uid,
                            issuer_key_id="0000000000000000",
                            signature_class="13x",
                            creation_date="1587731664",
                        ),
                        gpg.Signature(
                            issuer_uid=dcc_uid,
                            issuer_key_id="881685B5EE0FCBD3",
                            creation_date="1588064955",
                            signature_class="13x",
                        ),
                        gpg.RevocationSignature(
                            issuer_uid=dcc_uid,
                            issuer_key_id="881685B5EE0FCBD3",
                            creation_date="1588066811",
                            signature_class="30x",
                            reason="00",
                            comment="revsig\\nInvalid email",
                        ),
                    ),
                    key_capabilities={
                        gpg.KeyCapability.Sign(True),
                        gpg.KeyCapability.Sign(False),
                        gpg.KeyCapability.Certify(True),
                        gpg.KeyCapability.Certify(False),
                        gpg.KeyCapability.Encrypt(True),
                    },
                    key_type=gpg.KeyType.public,
                    uids=(uid,),
                    owner_trust="-",
                    sub_keys=(
                        gpg.SubKey(
                            key_type=gpg.KeyType.public,
                            key_id="AAAAAAAAAAAAAAAA",
                            fingerprint="BBBBBBBBBBBBBBBBBBBBBBBBAAAAAAAAAAAAAAAA",
                            validity=gpg.Validity.unknown,
                            key_length=2048,
                            pub_key_algorithm=1,
                            creation_date="1587731664",
                            key_capabilities={gpg.KeyCapability.Encrypt(False)},
                        ),
                    ),
                    origin="http://keyserver.dcc.sib.swiss:11371",
                )
            ],
        )
        self.assertEqual(
            keys[0].valid_signatures,
            [
                gpg.Signature(
                    issuer_uid=uid,
                    issuer_key_id="0000000000000000",
                    signature_class="13x",
                    creation_date="1587731664",
                )
            ],
        )

    @staticmethod
    def test_parse_keys_w_rev():
        key_payload = io.BytesIO(
            b"""tru::1:1576574206:0:3:1:5
pub:-:2048:17:CCCCCCCCCCCCCCC:1282220531:1592219853::-:::scESC::::::23::0:
fpr:::::::::CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC:
uid:-::::1534238253::DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD::Roundhouse Team <team@roundhouse.gov>::::::::::0:
sig:::17:CCCCCCCCCCCCCCC:1534238253::::Roundhouse Team <team@roundhouse.gov>:13x::CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC:::10:
sig:?::1:AAAAAAAAAAAAAAAA:1512502244::::[User ID not found]:13x::AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA:::8:
rev:?::1:BBBBBBBBBBBBBBBB:1521078619::::[User ID not found]:30x,00::BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB:::8:
"""
        )
        gpg.parse_keys(key_payload)

    @staticmethod
    def test_parse_keys_w_sub_sub():
        key_payload = io.BytesIO(
            b"""tru::1:1576574206:0:3:1:5
pub:-:2048:17:CCCCCCCCCCCCCCC:1282220531:1592219853::-:::scESC::::::23::0:
fpr:::::::::CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC:
uid:-::::1534238253::FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF::Roundhouse Team <team@roundhouse.gov>::::::::::0:
sub:-:2048:16:DDDDDDDDDDDDDDDD:1282220531:1592219869:::::e:::::::
fpr:::::::::DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD:
sig:::17:CCCCCCCCCCCCCCC:1534238269::::Roundhouse Team <team@roundhouse.gov>:18x::CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC:::10:
sub:-:4096:1:AAAAAAAAAAAAAAAA:1396950003:1704188403:::::s::::::23:
fpr:::::::::AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA:
sig:::17:CCCCCCCCCCCCCCC:1396950003::::Roundhouse Team <team@roundhouse.gov>:18x:::::10:
"""
        )
        gpg.parse_keys(key_payload)

    def test_extract_key_id(self):
        partial_msg = io.BytesIO(b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xff]\xdf")
        pos = partial_msg.tell()
        (key_id,) = gpg.extract_key_id(partial_msg)
        self.assertEqual(key_id, "F30D4C33407A3F63")
        self.assertEqual(partial_msg.tell(), pos)

        wrong_msg = io.BytesIO(b"\x04\x8c\x03\xf3\rL3@z?c\x01\x03\xff]\xdf")
        with self.assertRaises(gpg.GPGError):
            list(gpg.extract_key_id(wrong_msg))

        partial_msg = io.BytesIO(
            b"\x85\x01\x0c\x03\xd8\x92\xc4\x19\x17\xb2\x01\x15\x01\x08"
            b"\x00\xafZ\xd8\x83\xba"
            + 248 * b"\000"
            + b"c\x1d\x16\x85\x02\x0c\x032\xf9\xe2\xf8\rR!^\x01\x10\x00\x97c"
            + 504 * b"\000"
            + b"\x1a\xa2\xf2&\xc0\xe5\x85\x02\x0c\x03\xcc\xfa\xa7\xf4\x12\xb0m"
            b"\x83\x01\x10"
        )

        key_ids = list(gpg.extract_key_id(partial_msg))
        self.assertEqual(
            key_ids, ["D892C41917B20115", "32F9E2F80D52215E", "CCFAA7F412B06D83"]
        )

    def test_extract_key_id_ascii_armored(self):
        partial_msg = io.BytesIO(b"-----BEGIN PGP MESSAGE-----\r\n\r\n")
        with self.assertRaises(gpg.GPGError) as cm:
            list(gpg.extract_key_id(partial_msg))
        self.assertEqual(
            format(cm.exception), "ASCII armored encrypted data is not supported"
        )

    def test_extract_key_id_from_sig(self):
        sig = (
            b"\x89\x023\x04\x00\x01\x08\x00\x1d\x16!\x04\xa8\x05h*\x13\xa1\x08\xca"
            b"\xab\x07\xef\xd0k\xe5(\xb1&\xd3\xe13\x05\x02_\x05\xcfP\x00\n\t\x10k"
            b"\xe5(\xb1&\xd3\xe13\xf8\x8a\x10\x00\x97\xb8\x0bWU\x05B\xd5\x17\xba\xd2"
            b"\x9b\xe4o\xd4`\x86Id\x86\xc4m\xa6=\xf2\xf3\xc5\xec\x8b\x16?Q\xfe\x06"
            b"\xf9\xdc`\xfa\xe4\x9d~)\xcaYx\x16\x06t@\xc0\xc0\xe4\xd5\x94\x1a\x19"
            b"\xa5yL.r\xba\x01-\xe8\xc6\xaf \xa8\xc9\xbef\x0b:C\xa4\x83Ahn\x85R\xbe"
            b"\x81\x02\xb14\xce\x89\xef4\xa4\xb5\x96\xcc\x81t/x\x16_\xb1|\xcbb+3"
            b"\xf4`w_\x93\x8cl\xa0\x95$<&\x8fu<\n\xd1-9^\\\x98\x04\xa0F\x9fE\xeb{"
            b"\xc4\xda\x9f\x9f\xf0\xc9#>\xba\xb5\xa8\x99Od\xcdb\x9en\xa0\xb2\x1c\xd2"
            b"\xe8\xa8\x03k\x04o\xd5\x08\xb7W\x19UL\xa4\x14P\xba\x10}XS\x18\x86M"
            b"\xa3k\xdc\x0fO&\xe6J-\xca\x13\xac\xe8\xacb\xc5\x81x\xe3\xf0\x0b\xe1["
            b"\x1f\xc7\xd1b\x89\xb7\xe3E\x06\xa7n\x83\xe8\xc0.\x96\xd8\x19\x9d\xc3J"
            b"\xfcR\xc1%\x960\x95Y\n\x98`\xb1C(KAwO\xf6\xf5\xe1\xdcCk\x96j\xbbC\xd3?"
            b"\x05\x07r\xc1#\x97\x1d\x15~L\xb8HQ\xb4|=\x8dG\x987wN\xc8\xbfm\x889\xd5"
            b"\x9c\xfd\x90@#C\xef\xcf\x12\r\xc1\xdd\x1e\xd2\xc3\x97[\xad\xcc\x86\xbc"
            b"\xa2\xc6R\xce\xf6L\x8f5;\xe0\xc4|\xe2\xeb\xfe\xe5\xb0\xa2F\xc2\x8eX"
            b"\xab \xe7\xe3\xf3\xe9SN\x16K\x99\xb9S}\xa7\x02\xaa\xd6\x9c\xb3\xb8"
            b"\xcelO\xd3!/2r|\x06\x10\x88$L\x99\x15 \xff\xa7\x812\xb6\xf6I\xba\x10"
            b"\xfd\x94\xb6\x05:`*\xa3\xd0\xfb\xf8~\x0f\xedPx\t\x1fq\xe0\x8e\xf0\xcb"
            b"\xf1\xab\xa0yF\xc7iL\xba\xb7e\xfa\xba\xfb\x8e\xe5w\x8bF\xd3\x91\xc7G4"
            b"\x9c]\xcc\xb8\x10\x0c\xdc\x81\xc0_\x1e\x83c=x2\xd6\xa9:s\xc5\xf0\x04#"
            b"\x0c`\xa2\x0c\x95&\x08\xf5:\x12\xe0\x8a\xb6\x86\x9dE\xea\xa9\xd6\xff"
            b"\xe9~\x91_|\x14.\xcb\x14\xdd\xb9\tR\x1a)\x04\xaf\xf6\xb7\xd4(&%\xc4"
            b"\xe9\xea\xac\x00\x17\xe3Nf\xd6\xd1\xa0\xb9\x99\xe3[\xa1O?a\xd7\x06\x9fc"
        )
        key_id = gpg.extract_key_id_from_sig(sig)
        self.assertEqual(key_id, "6BE528B126D3E133")
        sig = b"""-----BEGIN PGP SIGNATURE-----

iQIzBAABCAAdFiEEqAVoKhOhCMqrB+/Qa+UosSbT4TMFAl8YCiwACgkQa+UosSbT
4TMEvQ/5AaSlYmpQK1MWQbNHiL4AMl55GfVlSnQBCpk3lsltOnycEVxP67jawpd7
2DSFBzcF5wRUiMo85R9YkNxBjFwfU2xlT738ZWFQIoOvvwyGJgOlgngZ7+BEGEqP
130pEIGeDVt6kzVuwqB09SZCATGR0qnkB/84+pFimN0tiFkFWHXLHgJ++uMO2Pg+
j45HbZzYh0qqKfY+X9lP5JIuCsnFQv5AOgnUFfy1vKNZd6OGCZuxCla5M+RpOcFD
F3w004RK2p3SNS7ikoATRSDnSHwiWDay8WesmVkOXQurevEONGdvUY91ZW41wA5T
HR3qHWZyg4fWy4G2azCbGA3AEnd2m+sWdvptpbL7InxutLET5hLEXNsnL1Ixiqr/
DViv+hBaixAT0LRCdci3G/FfWAC38zZjo/A6JLpCGaqTuWl1OWkDPxQXPNYT4Wgk
Y9YJFrgF1MQ2BrUuYzrC+ClEvO/TBsVe2r9l4ElZs3rZBH1vOxyCyuAHreNw46R6
MfwW9lYob56DUTRBVrhZI46/LQD/dlwtYedN5p/CNpzavuzQHAmJD1G22WZHwTqV
SL/bsOqwOqJP1BpQloAR+jEnP6jPbQAmPEn3xU+3IPWjKFe95QUj3PIeeSqst2Mt
xERXJ/PqFoAktjU/mhNeDHovXdB+7378YkY3NHNSlFytYG0D5jo=
=qg97
-----END PGP SIGNATURE-----
"""
        key_id = gpg.extract_key_id_from_sig(sig)
        self.assertEqual(key_id, "6BE528B126D3E133")
        key_id = gpg.extract_key_id_from_sig(sig.replace(b"\n", b"\r\n"))
        self.assertEqual(key_id, "6BE528B126D3E133")

    def test_capture_fpr(self):
        b = b"""
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [undefined]
[GNUPG:] VALIDSIG AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 2020-03-10 1583836167 0 4 0 1 8 00 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(
            gpg.capture_fpr(b), ["AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"]
        )
        b = b"""
gpg: Signature made Mi 12 Feb 2020 15:34:46 CET
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(gpg.capture_fpr(b), [])
        b = b"""
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 0
[GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [undefined]
[GNUPG:] VALIDSIG AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA 2020-03-10 1583836167 0 4 0 1 8 00 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
[GNUPG:] NEWSIG
gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
gpg:                using RSA key BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
[GNUPG:] KEY_CONSIDERED BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 0
[GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
[GNUPG:] KEY_CONSIDERED BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 0
[GNUPG:] GOODSIG 066E042415D8FD9C Gunnery Sgt. Hartmann <hartmann@bullshit.gov>
gpg: Good signature from "Gunnery Sgt. Hartmann <hartmann@bullshit.gov" [undefined]
[GNUPG:] VALIDSIG BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB 2020-03-10 1583836167 0 4 0 1 8 00 BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB
[GNUPG:] TRUST_UNDEFINED 0 pgp
gpg: WARNING: This key is not certified with a trusted signature!
gpg:          There is no indication that the signature belongs to the owner.
"""
        self.assertEqual(
            gpg.capture_fpr(b),
            [
                "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
                "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB",
            ],
        )


class IntegrationTest(unittest.TestCase):
    def _setup_key(self):
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        self.s.import_file(CN_PUB_KEY)
        (key,) = self.s.list_sec_keys()
        return key

    def setUp(self):
        gpg_homedir = tempfile.TemporaryDirectory()
        self.gpg_homedir = gpg_homedir
        d = gpg_homedir.__enter__()
        self.s = gpg.GPGStore(d)

    def tearDown(self):
        try:
            self.gpg_homedir.__exit__(None, None, None)
        except FileNotFoundError:
            pass


@unittest.skipUnless(WITH_GPG, "Integration tests")
class TestGPGIntegration(IntegrationTest):
    def test_revoke(self):
        if self.s._version < (2, 2, 0):  # pylint: disable=protected-access
            self.skipTest("Key revocation not supported for gnupg < 2.2.0")
        key = self._setup_key()
        self.assertTrue(key.validity != gpg.Validity.revoked)
        self.s.revoke(key.fingerprint, passphrase="test")
        (key,) = self.s.list_sec_keys()
        self.assertEqual(key.validity, gpg.Validity.revoked)

    def test_list_keys(self):
        self._setup_key()
        self.s.import_file(UMLAUTS_PUB_KEY)

        def extract_keys(**kwargs):
            keys = self.s.list_pub_keys(**kwargs)
            if keys[0].uids == (CN_UID,):
                key_1, key_2 = keys
            else:
                key_2, key_1 = keys
            return key_1, key_2

        pub_key, umlauts_key = extract_keys()

        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        self.assertFalse(pub_key.signatures)
        self.assertEqual(umlauts_key.uids, (UMLAUTS_UID,))
        self.assertEqual(umlauts_key.key_type, gpg.KeyType.public)
        self.assertFalse(umlauts_key.signatures)

        pub_key, umlauts_key = extract_keys(sigs=True)
        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        self.assertTrue(pub_key.signatures)
        self.assertEqual(umlauts_key.uids, (UMLAUTS_UID,))
        self.assertEqual(umlauts_key.key_type, gpg.KeyType.public)
        self.assertTrue(umlauts_key.signatures)

        keys = self.s.list_pub_keys(sigs=True, keys=(pub_key.fingerprint,))
        self.assertEqual(keys, [pub_key])

        (sec_key,) = self.s.list_sec_keys()
        self.assertEqual(sec_key.uids, (CN_UID,))
        self.assertEqual(sec_key.key_type, gpg.KeyType.secret)

        keys = self.s.list_pub_keys(sigs=True, keys=())
        self.assertEqual(keys, [])

    def test_gen_key(self):
        self.s.gen_key(
            full_name=CN_UID.full_name,
            email=CN_UID.email,
            passphrase="test",
            key_length=1024,
        )
        (pub_key,) = self.s.list_pub_keys()
        self.assertEqual(pub_key.uids, (CN_UID,))
        self.assertEqual(pub_key.key_length, 1024)
        self.assertEqual(pub_key.key_type, gpg.KeyType.public)
        (sec_key,) = self.s.list_sec_keys()
        self.assertEqual(sec_key.uids, (CN_UID,))
        self.assertEqual(sec_key.key_length, 1024)
        self.assertEqual(sec_key.key_type, gpg.KeyType.secret)

    def test_decrypt(self):
        message = b"a"
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        decrypted = io.BytesIO()

        def capture_output(stream):
            decrypted.write(stream.read())
            decrypted.seek(0)

        with self.assertRaises(BaseException):
            self.s.decrypt(
                source=io.BytesIO(ENCRYPTED_MESSAGE),
                output=capture_output,
                passphrase="wrongpass",
            )
        self.s.decrypt(
            source=io.BytesIO(ENCRYPTED_MESSAGE),
            output=capture_output,
            passphrase="test",
        )
        self.assertEqual(decrypted.read(), message)

    def test_encrypt_decrypt(self):
        key = self._setup_key()
        message = b"a"
        decrypted = io.BytesIO()

        def capture_output(stream):
            decrypted.write(stream.read())
            decrypted.seek(0)

        decrypt = partial(self.s.decrypt, output=capture_output, passphrase="test")
        self.s.encrypt(
            source=message,
            output=decrypt,
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
        )
        self.assertEqual(decrypted.read(), message)

    def test_encrypt_decrypt_compression(self):
        key = self._setup_key()
        message = b"a" * 50
        compressed = io.BytesIO()
        not_compressed = io.BytesIO()

        def capture_output(stream, buffer):
            buffer.write(stream.read())
            buffer.seek(0)

        self.s.encrypt(
            source=message,
            output=partial(capture_output, buffer=compressed),
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
            compress_algo=gpg.CompressAlgo.ZLIB,
            compress_level=9,
        )

        self.s.encrypt(
            source=message,
            output=partial(capture_output, buffer=not_compressed),
            recipients=(key.fingerprint,),
            trust_model=gpg.TrustModel.always,
            compress_algo=gpg.CompressAlgo.NONE,
        )

        self.assertLess(len(compressed.read()), len(not_compressed.read()))

    def test_delete_keys(self):
        self.s.import_file(SGT_HARTMANN_PUB_KEY)
        (key,) = self.s.list_pub_keys()
        self.assertEqual(key.fingerprint, SGT_HARTMANN_FPRT)
        self.s.delete_keys(key.fingerprint)
        self.assertFalse(self.s.list_pub_keys())

    def test_detach_sig(self):
        self.s.import_file(CN_SEC_KEY, passphrase="test")
        (cn_key,) = self.s.list_pub_keys()
        (cn_key,) = self.s.list_sec_keys()
        msg = b"Camelot! ... It's only a model."
        with self.s.detach_sig(msg, passphrase="test") as out:
            sig = out.read()
        fpr = self.s.verify(sig=sig, src=msg)
        self.assertEqual(fpr, cn_key.sub_keys[0].fingerprint)
        with self.assertRaises(gpg.GPGError):
            self.s.verify(sig=sig, src=b"Wrong message")


def to_info(key):
    return gpg.KeyInfo(
        uid=key.uids[0],
        fingerprint=key.fingerprint,
        key_algorithm=key.pub_key_algorithm,
        creation_date=key.creation_date,
        expiration_date=key.expiration_date,
        key_length=key.key_length,
    )


@unittest.skipUnless(KEY_SERVER is not None and WITH_GPG, "Key Server Tests")
class TestGPGKeyserver(IntegrationTest):
    def test_send_search(self):
        self.s.import_file(CN_PUB_KEY)
        self.s.import_file(UMLAUTS_PUB_KEY)
        key_a, key_b = self.s.list_pub_keys()

        self.s.send_keys(key_a.key_id, key_b.key_id, keyserver=KEY_SERVER)

        keys = gpg.search_keys(key_a.uids[0].full_name, keyserver=KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_a)])

        keys = gpg.search_keys(key_b.uids[0].full_name, keyserver=KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_b)])

        # By fingerprint
        keys = gpg.search_keys(key_b.fingerprint, keyserver=KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_b)])

        # By ID
        keys = gpg.search_keys(key_b.key_id, keyserver=KEY_SERVER)
        self.assertEqual(list(keys), [to_info(key_b)])

    def test_recv_keys(self):
        try:
            with tempfile.TemporaryDirectory() as gpg_temp_dir:
                temp_store = gpg.GPGStore(gpg_temp_dir)
                temp_store.import_file(SGT_HARTMANN_PUB_KEY)
                temp_store.import_file(UMLAUTS_PUB_KEY)
                key_a, key_b = temp_store.list_pub_keys(sigs=True)
                self.assertTrue(key_a.signatures or key_b.signatures)
                temp_store.send_keys(key_a.key_id, key_b.key_id, keyserver=KEY_SERVER)
        except FileNotFoundError:
            pass
        self.s.recv_keys(key_a.key_id, key_b.fingerprint, keyserver=KEY_SERVER)
        keys = set(self.s.list_pub_keys(sigs=True))
        self.assertEqual({key_a, key_b}, keys)


CN_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----


mI0EXcPQFwEEAMm1d9jlC5Zw1y/buUV6MuT5ABp4H8Sr3+MUf/GMFKPvd4Qboitd
Yf3eX753y0MNgzFlsSl8CtSO9pd+lqEt4WNMbIVWti04JAM6tD4lqeSzW57vlQlb
sGRf70fg9Qt1dc2LR0HcUi6bw46eWbSGGR3ZM7jbxB9F/VjpaTKI9S69ABEBAAG0
KkNodWNrIE5vcnJpcyA8Y2h1Y2subm9ycmlzQHJvdW5kaG91c2UuZ292PojOBBMB
CAA4FiEEKGahQqFvVOButzz5Lj5/4o/kEq4FAl3D0BcCGy8FCwkIBwIGFQoJCAsC
BBYCAwECHgECF4AACgkQLj5/4o/kEq59GQP+ONPvcLBccHAYbf0utUCiWZux9EwH
BN/TfOo1Vpgbq/DiBZcPoVeDvGTJDPg3dSUxsQzMHztp1UAi6o5Q6XCAGALMlxU9
QIHb4fyfbTQX1QrtB2EyLD/9O8nePrhqTqAd+xpdyvAw8L+hMoTyk0JnVW/UBVLH
RK4L9NaBxhSapfS4jQRdw9AXAQQAsuju5Ccwg3mDJcV+bR9yHIc+XyXcsbzFW0VU
lgNXiQoVHKJQ+X/SgElmxKwik8yeQ4CiM3/GeKmjhKB7JZnp4U+61Dfew+qc6Net
uAdNu7f26x5q237tjU9zx4JzhdnDMhrXkuYl/VR95S7CHlt1QlVxP1/Psx11a/Ej
xB+7ZDsAEQEAAYkBawQYAQgAIBYhBChmoUKhb1Tgbrc8+S4+f+KP5BKuBQJdw9AX
AhsuAL8JEC4+f+KP5BKutCAEGQEIAB0WIQRK0g8wPkscJLLIn7rzDUwzQHo/YwUC
XcPQFwAKCRDzDUwzQHo/YzIkA/4m5TWu7yvVzxjtybqHpShxrrZeTzWRrSu/SiCQ
bR8pGWP9NhnRDcfH/zMX6sMMorE6BBHI8fFMzJB974Lv0q2S8njVMuv+tIH7t+Um
KxlTY2Nz+uzFjuv2rDvPL4ozb4iZfOFq1WOwrml3fpE3qQzqLypd4OBj4H+ITghJ
DiJ55n2eBACItELQ3PcTXeh7XmXTeV3/lVhGHWlhUvcQUCNZCo9AFoX9aYBPX6OF
arsEzjNcZilc/R+r3xYF4kV2kspQQPGoslHN0Zq4Wxp0tOh/IevqjJ3bSzcCqwGz
c9fhdV81kJHiCZUFT03X+D5LrOJJ857/vwL1Roi5UgjRjX1jMFV/tg ==
=0f3v
-----END PGP PUBLIC KEY BLOCK-----
"""

CN_SEC_KEY = b"""-----BEGIN PGP PRIVATE KEY BLOCK-----
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 ==
=G52m
-----END PGP PRIVATE KEY BLOCK-----
"""


UMLAUTS_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EXcQPZwEEAMFLJh5nhUL3e2cIIEHw2lUwleWu/FMgy5GVVGho1pzSjqxc2cML
0PTLePd5/w/BOHH7tx2404CcwNVF8NrdPgFLV7mZu0Q4eUJNG+yz4EpUBCVK2si/
vNijeF3YH2vO18cyTCGiCW6vXi1cJqf6k73PCrSOPqj7QuNi41Ahkr5FABEBAAG0
KVdpdGggVW1sw6R1dHMgPHdpdGgudW1sYXV0c0B1bWxhdXRzLmluZm8+iM4EEwEI
ADgWIQS1Rlqvu4FLhxmgfbVt8x+lbBrrcQUCXcQPZwIbLwULCQgHAgYVCgkICwIE
FgIDAQIeAQIXgAAKCRBt8x+lbBrrca0uA/9Dj+s3F4FzO2MsDGaD8WXuZ8LPFSRW
zehWewIeRS6hT7L9+LNF3m0GVOueXCPrHj/8iBstc5UbFuph3Pu/3Cgw5YnnSkcC
/LPFGRlc5Z6oRF259qZNW8NafOoi1E4o/LAN1QG5SChYUS2jcyDrosJekGEaPP3n
0KqQQPYC7Smu3riNBF3ED2cBBADClqvryRmApeYO8tQcGWaiFa27vFmM4FCAzbPK
tt88N1GaHWgycgVV/dChhWDVVkEDnas+WwVMz39ZZi8tejBrk1mtoIIBctn8AWJL
W7VH9ZOrfPm1UuqLE7YMhu9hl7i4ANIJnJMsjYnOnb/Z9ButzIobZdigV7a1j47t
pXHnuwARAQABiQFrBBgBCAAgFiEEtUZar7uBS4cZoH21bfMfpWwa63EFAl3ED2cC
Gy4AvwkQbfMfpWwa63G0IAQZAQgAHRYhBHW94ZPbXFu5K4+7zYDWfVUAnJDMBQJd
xA9nAAoJEIDWfVUAnJDMBb4EAKt1sMe/6MhTsn3weshU5U6T217W/lxXmz5vErnO
yd0WaMWOsGN7WEV/UtoVMZ/EBBdIqswUiujDPd8oL3yCsNBZNDigyCGX6shqCeRP
AAUkpC/0KWDP33PNenGSMFzoBcghYBRUuVucvamzskDfidsKbLv+/7FkRwzstg7L
5/GUKGkD/2LlSZlm+qDBzDt5q+tQaGOQnJmebF7o4FQ8feY7GNcFQ9DMV9lR5vuZ
jV4OTQVS2Prv2ASuN7rPRbUYeJD5t7Mko9DzL+6XGgjPim6Wg1xlRSRkskksbX3y
QndrRiimeh43nPBLOBRN9HQoaywkc6hoAZTQAvCgOZli507dp8Mz
=sfjX
-----END PGP PUBLIC KEY BLOCK-----
"""

SGT_HARTMANN_UID = gpg.Uid(
    full_name="Gunnery Sgt. Hartmann", email="hartmann@bullshit.gov"
)

SGT_HARTMANN_FPRT = "DA1A363BC5E11DA73806A8A4E3E4024A0E56221E"

SGT_HARTMANN_PUB_KEY = b"""-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBF3MKrABCACw6ZFUve5lTSCtDfhcuDH0pFd7AF0uM1o4kEBLKfcEImrgGo7S
LX7tooxsx2gNv4ARHW4XL5pixLj4itRBDeUGczdDzNwQeE6HmWNW2JtqF7/h1TTr
iw3UwS3eOG8sNrZxx2yLoIF/JZP2CDLz7jXDhuAvLdoK8hzNX82PVUkoVKtPYhbk
1gw3y1coH2Z6rdAp6EwrMSg95vbwzKjuhcyYlxRodwlvmJhDBRxr86QCCRNiA6R8
7Nugtt/wP5XbxcxgNBmciX10NPR4eIoQk4acyGlDvOdrF50h2DPAw3HUdOw513xN
4b7NOHlJJk0+Uo4yKcZ2LXmeiSNU6TY93moLABEBAAG0LUd1bm5lcnkgU2d0LiBI
YXJ0bWFubiA8aGFydG1hbm5AYnVsbHNoaXQuZ292PokBVAQTAQgAPhYhBNoaNjvF
4R2nOAaopOPkAkoOViIeBQJdzCqwAhsDBQkDwmcABQsJCAcCBhUKCQgLAgQWAgMB
Ah4BAheAAAoJEOPkAkoOViIeSIEH/RDXmilCOkoR6uxJgNm7MUqavL2/s3ks57uk
qDrldJm3T5SKBCgrUSP3UaiBIwq+M32k23+CLNl8RQBVwhFAnJcHmMK/fEVcSK8G
HKy4CS8yu1xwDIgWWeDBtbo6r4IEah19dBSLOrxeyBFoI65pRHP18AWMaHnKv5nz
GXdGiLlRl6rKbyBxnesXweR+IIZJhciphhzuV8HtnyHsess0Toq2kjiCGp7Kjkz1
6YQSiHawAj3nV+/RK0kNLZsuQOgY7X2WVw4HAU//Ap1XdtMbOxberLcUU7Rzeudg
uiWfkmTMM36mhojOjtPhgB1J9ZZZCe5KM7hRB2VFuJE45r3S5LCIswQTAQgAHRYh
BChmoUKhb1Tgbrc8+S4+f+KP5BKuBQJdzCsMAAoJEC4+f+KP5BKuz3oD/inKIFnX
KzFblhnLeHf+tW7JQ9EVu+7rTKKyHHay7sgzVQ2KfySemFYwn2fgXUOuZtuk0axe
BHV0qTgL0vfnadxXF96tjFQ14xhBym6KR2ccvQN+/GLM5c0iBbp59azMy0q4VyJH
5c435C3Yq9ownEvJRRdYdJgaFTU1wjwQs7qZuQENBF3MKrABCADguI7FCLGtG6rZ
/EJv58Bnihfz9OuUSc/R6DaFbGGtolTzpTskX3pK1daYigPDDv4iMN5KnO1eeJyp
wmyRpC2yKBLw70z87QGPSviQHWTvyj22YF9ByywFKAuZeY2tpWwkLW2bbVm38LLT
hwmDRkQ8nQZPufvpDOHq6AC84ymW5TuHKEYTfF//14hyonOK40HNjSdEC6ZvfN2n
N80y9g3IvaNGsPOo3MW4z877W9v/IAvFlTEVnyRA0hXEykFnUYk//Eh6cnDTlrCO
hLwd9bbcivPPpjiGYLjgvDghyXK/qv2T4ahiMq2qxp5F8c4x/aTq8U//9GIEU7nd
danxrHsZABEBAAGJATwEGAEIACYWIQTaGjY7xeEdpzgGqKTj5AJKDlYiHgUCXcwq
sAIbDAUJA8JnAAAKCRDj5AJKDlYiHiyUB/9ZH1ZUF/lHsFjFgypyv1BfqFIyYZk9
1/zSHdwUl7svtBrc8XewisxgCBywbzJWHCooAAkSmwSrCFnqVfoUFCdAYxd3lrX1
DA2WEDcIr3Im2wUOKw0fB5yyBcep5QAzoRFl3uZwOcAYrqq38L7xAoRka991oPIk
nzY4GfcR+7ftmzKDPcXjTIbLQwmDNR0vDy7kNBB7ycHWwrMfadWGrGmeQ8pgTnP4
Egfkuv5fEQ3njiV3QcRwmTtqPEVPmA1G/TStAPZa9cBfYHTxH1JzxwVhYvBWXoA3
vwMVdh5kI53+JxM0k7OS6SI8cpf4pmiyLypdkyOCql2HYe7iPuRk0IYi
=lmDh
-----END PGP PUBLIC KEY BLOCK-----
"""

ENCRYPTED_MESSAGE = (
    b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xff\\\t\xea;f\xb4`b\x05\xfeJ(\xdf"
    b'^`"\x87B/\x8b\xdfYD\xb4^\x03\x99R\x0bn\x1c\x12\xca6Ju\x8e\x03\x1fI'
    b"U|~\x89>w\x18'\xad\xbb\xa0\x86\x85\x1a\r\x93\xe8$\xd0H\xf8\xef\xfa"
    b"\x16\x87\x17K|\x9d$\xeb<\xbd\x0e3N\xbd{\xf1bT\xcd\xa3-\xc6\x83\x0e"
    b"\x96\xc5\xcb--\x93P\x15\xe7@\x84\x8fR\xfa\t\xc2\x99\x1f\xc4\xf0`\xfc"
    b"\x87v\x04\x1c\x9ef\x03\xc7Jh\x840\x112/\x1b\x06L\xf4\xd2<\x01\xbb"
    b"\xd2\x80=\x17\x9d\xc5ww\n\xd9\xf7%\x9fpzg\\\x1fz\x02\xf8$\x17\xe7V"
    b"\x825\xdd\n\xedu\x12\xa9\xac\xab5\x99\x9b\x7f\x9a\x8f\\\xd1\xa8\xb2"
    b"\xe9!\x1dH\xa3\x1c\xd9\xf9\xd1l\x86\xe7\x12"
)
ENCRYPTED_MESSAGE_CN = (
    b"\x84\x8c\x03\xf3\rL3@z?c\x01\x03\xffw\x058f\x95\xbd\xe5\x8d1q"
    b"\xdd\x80\xbd\x8d\xb5v5\xb3\xcf\x92\xb9?6\x8a \xfc\xc9\xc9\xad"
    b"\xccI\xd7\xc1@\xe3\xb3]\xbag\xcb\xae\x99\xbdkB\xdd\xcaS\x86598"
    b"\x00\xdd\xe1C\xad$1<\xe0\xac\xd5\x9b4\xc0w\xb5O\xf6\x8a/\x88\x7f"
    b"\x1b!\xb7\x08x\x0cUV\xd3\x1e\xaba\xe8\x03]\xc79\xea\x05\x0f*\xdc"
    b"\xa5\x88\xac\xa3Y\xa0\xfb\xd24\xf4\xa0hD\xe3\xaa\x11\xd1\xf9o"
    b"\xf7\xf7\xc3\xc8:\x0c\x16\xd5W\xfa%\x8f7\xd2F\x01q\x07\xd0\xde"
    b"~\xe1V\x00h\xc0O=ls?\x13\xfc\x9e\x8aY>e\x10$\xd2\xdb;\x18\xc0"
    b"\xf7]\xa2)r\xfe\xf8\xd9\x05\xcb\xcc\r8\xb5\xaf\x8b\xa0\xff.\xa1T"
    b"&_\x17\xdb\x0bi\x04\xd0\xa3\xa2\x9c\x8b\x80\x96\xb5\x10I\xea`"
)



import unittest

from gpg_lite import parser


class TestParser(unittest.TestCase):
    def test_expand_token_sequence(self):
        seq = (1, 2, 3)
        self.assertEqual([seq], list(parser.expand_token_sequence(seq)))
        seq = ((1, 2), 2, 3)
        self.assertEqual({(1, 2, 3), (2, 2, 3)}, set(parser.expand_token_sequence(seq)))
        seq = (1, (2, 3), (2, 3, 4))
        self.assertEqual(
            {(1, i, j) for i in (2, 3) for j in (2, 3, 4)},
            set(parser.expand_token_sequence(seq)),
        )

    def test_expand_token_sequences(self):
        seqs = {(0, 2, 3): 1, (1, (2, 3), (2, 3, 4)): 2}
        ref_seqs = {(1, i, j): 2 for i in (2, 3) for j in (2, 3, 4)}
        ref_seqs[(0, 2, 3)] = 1
        self.assertEqual(parser.expand_token_sequences(seqs), ref_seqs)



import io
import platform
import os
import re
import tempfile
import base64
from contextlib import contextmanager
from enum import Enum, auto
from typing import (
    Optional,
    List,
    Tuple,
    AbstractSet,
    Iterator,
    Iterable,
    Union,
    Collection,
    Sequence,
    Any,
    Type,
    IO,
    Dict,
    ContextManager,
    Callable,
)
from urllib.error import HTTPError
from urllib.parse import unquote, quote
import urllib.request

from dataclasses import dataclass, fields as dataclass_fields, field

from .parser import compile_grammar
from .deserialize import deserialize
from .cmd import (
    cmd as _cmd,
    cmd_devnull,
    cmd_pipe,
    cmd_pipe_stdout,
    expect,
    GPGError,
    stderr_lookahead,
)


@dataclass(frozen=True)
class Uid:
    """GPG User ID representation.

    The GPG user ID format can contain user's name and/or email address in the
    following formats:

    - Alice (Alice's test PGP key) <alice@example.com>
    - Alice
    - alice@example.com
    """

    full_name: Optional[str]
    email: Optional[str]

    def __post_init__(self):
        if self.full_name is None and self.email is None:
            raise ValueError("Both 'full_name' and 'email' cannot be None.")

    def __repr__(self):
        if self.full_name is not None and self.email is not None:
            return f"{self.full_name} <{self.email}>"
        return self.full_name or self.email

    @staticmethod
    def from_str(s):
        if s is None or s[0] == "[" and s[-1] == "]":
            # Something like "[User ID not found]" (can be translated into other languages)
            return None
        if re.search(r" <\S+>", s) is not None:
            return Uid(*s.rstrip(">").split(" <"))
        if "@" in s:
            return Uid(full_name=None, email=s)
        return Uid(full_name=s, email=None)


class KeyType(Enum):
    """Representation of the GPG key type. Only two type of keys exist:
    public and secret (often also referred to as 'private')"""

    public = auto()
    secret = auto()


class TrustModel(Enum):
    pgp = "pgp"
    classic = "classic"
    tofu = "tofu"  # Trust on first use.
    tofu_pgp = "tofu+pgp"
    direct = "direct"
    always = "always"
    auto = "auto"


class Validity(Enum):
    unknown_new = "o"  # Unknown (this key is new to the system)
    invalid = "i"  # The key is invalid (e.g. missing self-signature)
    disabled = "d"  # The key has been disabled
    #                  # (deprecated - use the 'D' in field 12 instead)
    revoked = "r"  # The key has been revoked
    expired = "e"  # The key has expired
    unknown = "-"  # Unknown validity (i.e. no value assigned)
    undefined = "q"  # Undefined validity.  '-' and 'q' may safely be
    #                  # treated as the same value for most purposes
    not_valid = "n"  # The key is not valid
    marginal_valid = "m"  # The key is marginally valid.
    fully_valid = "f"  # The key is fully valid
    ultimately_valid = "u"  # The key is ultimately valid. This often means
    #                       # that the secret key is available, but any key
    #                       # may be marked as ultimately valid.
    well_known_private_part = "w"  # The key has a well known private part.
    special_validity = "s"  # The key has special validity.  This means
    #                              # that it might be self - signed and
    #                              # expected to be used in the STEED system.


class SignatureValidity(Enum):
    good = "!"
    bad = "-"
    no_public_key = "?"
    error = "%"


class CompressAlgo(Enum):
    ZLIB = "zlib"
    ZIP = "zip"
    BZIP2 = "bzip2"
    NONE = "uncompressed"


key_capabilities = {}


def key_capability(key: str):
    def decorator(cls):
        wrapped = dataclass(frozen=True)(cls)
        for variant in {key, key.upper()}:
            key_capabilities[variant] = wrapped(variant.isupper())
        return wrapped

    return decorator


@dataclass(frozen=True)
class KeyCapabilityBase:
    global_: bool


class KeyCapability:
    @key_capability("e")
    class Encrypt(KeyCapabilityBase):
        pass

    @key_capability("s")
    class Sign(KeyCapabilityBase):
        pass

    @key_capability("c")
    class Certify(KeyCapabilityBase):
        pass

    @key_capability("a")
    class Authentication(KeyCapabilityBase):
        pass

    @key_capability("?")
    class Unknown(KeyCapabilityBase):
        pass

    @key_capability("D")
    class Disabled(KeyCapabilityBase):
        pass


@dataclass(frozen=True)
class Signature:
    issuer_uid: Optional[Uid]
    issuer_key_id: str
    signature_class: str
    creation_date: str
    expiration_date: Optional[str] = None


@dataclass(frozen=True)
class RevocationSignature(Signature):
    reason: Optional[str] = None
    comment: Optional[str] = None


@dataclass(frozen=True)
class SubKey:
    key_id: str
    key_type: KeyType
    fingerprint: str
    key_length: int
    pub_key_algorithm: int
    creation_date: str
    validity: Validity = Validity.unknown
    expiration_date: Optional[str] = None
    key_capabilities: AbstractSet[KeyCapabilityBase] = frozenset()
    signatures: Tuple[Signature, ...] = ()

    @property
    def valid_signatures(self) -> List[Signature]:
        rev_sig_ids = {
            s.issuer_key_id
            for s in self.signatures
            if isinstance(s, RevocationSignature)
        }
        return [s for s in self.signatures if s.issuer_key_id not in rev_sig_ids]


@dataclass(frozen=True)
class Key(SubKey):
    uids: Tuple[Uid, ...] = field(default=(), metadata=dict(required=True))
    owner_trust: str = "-"
    sub_keys: Tuple[SubKey, ...] = ()
    origin: Optional[str] = None

    def __post_init__(self):
        """Raise error if there is any required field with no value."""
        missing_fields = [
            f.name
            for f in dataclass_fields(type(self))
            if f.metadata.get("required", False) and not getattr(self, f.name)
        ]
        if missing_fields:
            l = len(missing_fields)
            plural = "s" if l > 1 else ""
            fields = ", ".join(f"'{f}'" for f in missing_fields)
            raise TypeError(f"__init__ missing {l} required argument{plural}: {fields}")


@dataclass(frozen=True)
class KeyInfo:
    uid: Uid
    fingerprint: str
    key_algorithm: int
    key_length: int
    creation_date: str
    expiration_date: Optional[str]


def parse_key_capabilities(s: Optional[str]):
    if s is None:
        return frozenset()
    return frozenset(key_capabilities[c] for c in s)


@dataclass
class ColonRecordBase:
    """According to [gnupg.git]/doc/DETAILS"""

    validity: Optional[Validity] = None  # 2
    key_length: Optional[int] = None  # 3
    pub_key_algorithm: Optional[int] = None
    key_id: Optional[str] = None  # 5
    creation_date: Optional[str] = None
    expiration_date: Optional[str] = None
    cert_S_N_uid_hash_trust_signature_info: Optional[str] = None
    owner_trust: Optional[str] = None
    user_id: Optional[str] = None  # 10
    signature_class: Optional[str] = None
    key_capabilities: AbstractSet[KeyCapabilityBase] = field(
        default=frozenset(), metadata=dict(deserialize=parse_key_capabilities)
    )
    issuer: Optional[str] = None
    flag: Optional[str] = None
    token_S_N: Optional[str] = None  # 15
    hash_algorithm: Optional[str] = None
    curve_name: Optional[str] = None
    compliance_flags: Optional[str] = None
    last_update: Optional[str] = None
    origin: Optional[str] = None  # 20
    comment: Optional[str] = None

    def into(self):
        return self


unimplemented_col_rec_ids = (
    "crt",  # X.509 certificate
    "crs",  # X.509 certificate and secret key available
    "uat",  # User attribute (same as user id except for field 10).
    "rvs",  # Revocation signature (standalone) [since 2.2.9]
    "pkd",  # Public key data [*]
    "grp",  # Keygrip
    "rvk",  # Revocation key
    "tfs",  # TOFU statistics [*]
    "tru",  # Trust database information [*]
    "spk",  # Signature subpacket [*]
    "cfg",  # Configuration data [*]
)

col_rec_types = {}


def col_rec_type(identifier):
    def decorator(cls):
        col_rec_types[identifier] = cls
        return cls

    return decorator


def extract_key_fields(rec):
    fields = (
        "validity",
        "key_length",
        "pub_key_algorithm",
        "key_id",
        "creation_date",
        "expiration_date",
        "owner_trust",
        "key_capabilities",
        "origin",
    )
    return {
        key: val
        for key, val in ((f, getattr(rec, f, None)) for f in fields)
        if val is not None
    }


@col_rec_type("sec")
class ColonRecordSec(ColonRecordBase):
    """Secret key"""

    def into(self):
        return Factory(key_type=KeyType.secret, **extract_key_fields(self))


@col_rec_type("pub")
class ColonRecordPub(ColonRecordBase):
    """Public key"""

    def into(self):
        return Factory(key_type=KeyType.public, **extract_key_fields(self))


@col_rec_type("sub")
class ColonRecordSub(ColonRecordBase):
    """ Subkey (secondary key) """

    def into(self):
        return SubKeyFactory(key_type=KeyType.public, **extract_key_fields(self))


@col_rec_type("ssb")
class ColonRecordSsb(ColonRecordBase):
    """ Subkey (secondary key) """

    def into(self):
        return SubKeyFactory(key_type=KeyType.secret, **extract_key_fields(self))


@col_rec_type("uid")
class ColonRecordUid(ColonRecordBase):
    """User id"""

    def into(self):
        return Uid.from_str(self.user_id)


@col_rec_type("fpr")
class ColonRecordFpr(ColonRecordBase):
    """Fingerprint"""

    def into(self):
        if self.user_id is None:
            raise ValueError("Field 'user_id' missing in record")
        return self.user_id


@col_rec_type("sig")
@dataclass
class ColonRecordSig(ColonRecordBase):
    """Signature"""

    # Here we ignore the change of type in the field. This type is
    # for parsing only
    validity: Optional[SignatureValidity]  # type: ignore

    def into_dict(self):
        uid = Uid.from_str(self.user_id)
        return dict(
            issuer_uid=uid,
            issuer_key_id=self.key_id,
            creation_date=self.creation_date,
            expiration_date=self.expiration_date,
            signature_class=self.signature_class,
        )

    def into(self):
        return Signature(**self.into_dict())


@col_rec_type("rev")
@dataclass
class ColonRecordRev(ColonRecordSig):
    """Revocation Signature"""

    def into(self):
        sig = self.into_dict()
        sig_class = sig.pop("signature_class")
        try:
            sig_class, revocation_reason = sig_class.split(",")
        except ValueError:
            revocation_reason = None
        return RevocationSignature(
            comment=self.comment,
            signature_class=sig_class,
            reason=revocation_reason,
            **sig,
        )


def parse_data_class(T: type, args: Sequence[Any]):
    fields = (f.metadata.get("deserialize", f.type) for f in dataclass_fields(T))
    return T(*(deserialize(f_type)(arg) for f_type, arg in zip(fields, args)))


def lex_colon(payload: IO[bytes]) -> Iterator:
    """Tokenizes the output of a call to the gpg command to list keys (payload)
    one line at a time. Works for both commands to list public or secret keys.
    Example of payload:
        tru::1:1574434353:0:3:1:5
        pub:u:4096:1:12477DE52DAD86CC:1574434343:::u:::escaESCA:
        fpr:::::::::621A2A449F234A0E9246CF8212477DE52DAD86CC:
        uid:u::::1574434343::9F08B014A8964D2FCA0B04A7B11CF5FDFD12DF13::bob <bob@example.com>:
        sub:u:4096:1:CFAF4D7B40DAF86B:1574434343::::::esa:
        fpr:::::::::39D91E3C8C9C2C0677F1A74BCFAF4D7B40DAF86B:

    :param payload: stream of bytes corresponding to the output of the gpg
        command.
    :return: generator of "token" that are either a "Factory" object or string.
    :raises ValueError:
    """
    # Loop through the output of the GPG command one line at a time.
    for rec in payload:
        rec = rec.rstrip(b"\n")
        if not rec:
            continue

        # Split line by ':' separator. The first element of the line is the
        # "type of record" (e.g. pub = public key, sub = sub key)
        t_id, *args = map(
            lambda s: s if s else None, rec.decode("utf-8", "replace").split(":")
        )
        args = args[: len(dataclass_fields(ColonRecordBase))]
        if t_id in unimplemented_col_rec_ids:
            continue
        try:
            token = parse_data_class(col_rec_types[t_id], args).into()
        except BaseException as e:
            raise ValueError(
                f"Error while parsing row of type {t_id}"
                f"\n{format(e)}"
                f"\ncolumns: {args}"
            ) from e
        if token is not None:
            yield token


class Factory:
    type_: Type[SubKey] = Key

    def __init__(self, origin: Optional[str] = None, **kwargs):
        self.kwargs = kwargs
        if origin:
            origin = origin.replace("\\x3a", ":").lstrip("0123456789 ")
            self.kwargs["origin"] = origin or None

    def append(self, key, val):
        self.kwargs[key] = self.kwargs.get(key, ()) + (val,)
        return self

    def append_subkey(self, val):
        if self.kwargs["key_type"] != val.key_type:
            raise ValueError(f"Wrong sub key type '{val.key_type}'")
        return self.append("sub_keys", val)

    def set(self, key, val):
        if key in self.kwargs:
            raise ValueError(f"Key '{key}' already set")
        self.kwargs[key] = val
        return self

    def build(self):
        try:
            return self.type_(**self.kwargs)
        except (TypeError, ValueError) as e:
            raise type(e)(e.args[0].replace("__init__", self.type_.__name__))


class SubKeyFactory(Factory):
    type_ = SubKey


# Factory, type : rule how to transform subsequences of tokens
# into reduced token sequences. For example, in:
#  -> (Factory, str): lambda fac, fpr: [fac.set("fingerprint", fpr)]
# the rule (lambda function) defines how to add the string (fingerprint) to
# the Factory (i.e. the main key object).
key_parser = compile_grammar(
    {
        ((Factory, SubKeyFactory), Uid): lambda fac, uid: [fac.append("uids", uid)],
        ((Factory, SubKeyFactory), str): lambda fac, fpr: [fac.set("fingerprint", fpr)],
        ((Factory, SubKeyFactory), (Signature, RevocationSignature)): lambda fac, sig: [
            fac.append("signatures", sig)
        ],
        (Factory, SubKey): lambda fac, sub: [fac.append_subkey(sub)],
        ((Factory, SubKeyFactory), list): lambda fac, lst: [fac.build(), lst],
        (SubKeyFactory, SubKey): lambda fac, key: [fac.build(), key],
        (Key, list): lambda key, lst: [[key] + lst],
    },
    list,
)


def parse_keys(payload: IO[bytes]):
    return key_parser(list(lex_colon(payload)) + [[]])


def parse_search_keys(source: IO[bytes]) -> Iterator[KeyInfo]:
    """According to https://tools.ietf.org/html/draft-shaw-openpgp-hkp-00"""
    try:
        b_line = next(source).rstrip(b"\r\n")
    except StopIteration:
        return
    if not b_line.startswith(b"info:"):
        raise ValueError(
            "Unknown response format from gpg, starting with\n"
            + b_line.decode("utf-8", "replace")
        )
    for b_line in source:
        line = b_line.decode("utf-8", "replace").rstrip("\r\n")
        if not line:
            continue
        try:
            (
                hdr_key,
                fingerprint,
                algorithm,
                key_length,
                creation_date,
                expiration_date,
                _,
            ) = line.split(":")
        except ValueError as e:
            raise ValueError(f"Wrong format for search_key: {line}") from e
        try:
            hdr_uid, uid, uid_creation_date, uid_expiration_date, _ = (
                next(source).decode("utf-8", "replace").rstrip("\n").split(":")
            )
        except StopIteration as e:
            raise ValueError("Unexpected end of source. Expected 'uid:' line") from e
        if hdr_key != "pub":
            raise ValueError(
                f"Unknown response format from gpg: "
                f"expected 'pub:', got '{hdr_key}'"
            )
        if hdr_uid != "uid":
            raise ValueError(
                f"Unknown response format from gpg: "
                f"expected 'uid:', got '{hdr_uid}'"
            )
        if creation_date and uid_creation_date and creation_date != uid_creation_date:
            raise ValueError(f"Key {fingerprint}: Mismatch in creation date")
        if (
            expiration_date
            and uid_expiration_date
            and expiration_date != uid_expiration_date
        ):
            raise ValueError(f"Key {fingerprint}: Mismatch in expiration date")
        yield KeyInfo(
            uid=Uid.from_str(unquote(uid)),
            fingerprint=fingerprint,
            key_algorithm=int(algorithm),
            creation_date=creation_date,
            expiration_date=expiration_date or None,
            key_length=int(key_length),
        )


def normalize_fingerprint(fp: str):
    valid_len = (8, 16, 32, 40)
    s = re.sub(r"^0[xX]", "", re.sub(r"\s", "", fp))
    if len(s) not in valid_len:
        raise ValueError("Invalid fingerprint length. " f"Allowed lengths {valid_len}")
    try:
        int(s, 16)
    except ValueError as e:
        raise ValueError(
            "Fingerprint contains invalid characters. " "Allwed characters: 0-9A-Fa-f"
        ) from e
    return f"0x{s}"


GPG_DEFAULT_DIR_BY_OS: Dict[str, Tuple[str, ...]] = {
    "Linux": (".gnupg",),
    "Darwin": (".gnupg",),
    "Windows": ("AppData", "Roaming", "gnupg"),
}


def get_default_gpg_config_dir() -> str:
    """Path of the directory where GnuPG stores the user's keyrings.
    The location of this directory is platform dependent.
    """
    return os.path.join(
        os.path.expanduser("~"), *GPG_DEFAULT_DIR_BY_OS[platform.system()]
    )


class GPGStore:
    """Main class providing functionality. Representation of a GPG 'home'
    directory that contains the user's keyrings
    """

    def __init__(self, config_dir: Optional[str] = None):
        self.config_dir = config_dir
        self._gpg_bin = None
        for _bin in ("gpg", "gpg2"):
            self._gpg_bin = _bin
            try:
                cmd_devnull((_bin, "--version"))
                break
            except FileNotFoundError:
                continue
        else:
            raise FileNotFoundError("gpg executable not found")
        self._version = self.version()
        self._pw_args_t: Tuple[str, ...] = ("--passphrase-fd", "0")
        if self._version >= (2, 2, 0):
            self._pw_args_t = ("--pinentry-mode", "loopback") + self._pw_args_t
        else:
            # Specific to older versions of gpg: the --batch option must be
            # added in order for the --passphrase-fd option to be used.
            self._pw_args_t = ("--batch",) + self._pw_args_t

    def _pw_args(self, pw):
        if pw is not None:
            return self._pw_args_t
        return ()

    @property
    def _base_args(self):
        args = (self._gpg_bin,)
        if self.config_dir is not None:
            args += ("--homedir", self.config_dir)
        return args

    def default_key(self) -> Optional[str]:
        config_file = os.path.join(
            self.config_dir or get_default_gpg_config_dir(), "gpg.conf"
        )
        try:
            with open(config_file) as f_cfg:
                default_key = next(
                    line for line in f_cfg if line.startswith("default-key ")
                )
            return default_key.split(" ")[1].rstrip("\r\n")
        except (FileNotFoundError, StopIteration):
            try:
                return self.list_sec_keys()[0].fingerprint
            except IndexError:
                return None

    def list_pub_keys(
        self, sigs: bool = False, keys: Optional[Collection[str]] = None
    ) -> List[Key]:
        return self._list_keys(option="--list-public-keys", sigs=sigs, keys=keys)

    def list_sec_keys(self, keys: Optional[Collection[str]] = None) -> List[Key]:
        return self._list_keys(option="--list-secret-keys", keys=keys)

    def _list_keys(
        self, option: str, sigs: bool = False, keys: Optional[Iterable[str]] = None
    ) -> List[Key]:
        """Ask gpg to list keys with the selected options.

        :param option: argument to pass to gpg, e.g. "--list-public-keys" to
            list the public keys in the user's keyring.
        :param sigs: if True, show signatures appended on keys.
        :param keys: key name, key ID or fingerprint to search for. If an
            empty list is passed, no key is returned.
            If keys is None, return all keys.
        :return: list of keys matching the search criteria.
        :raise GPGError:
        """
        if keys is None:
            keys = ()
        else:
            keys = tuple(keys)
            if not keys:
                return []
        # Add arguments that are always required for gpg to produce the
        # correct output. Note that for gpg versions < 2.1, the
        # --with-fingerprint option is required to be passed twice. If passed
        # only once the fingerprint of the subkeys are not printed.
        args: Tuple[str, ...] = (
            option,
            "--fingerprint",
            "--fingerprint",
            "--with-colons",
        )
        if sigs:
            args += ("--with-sig-list",)
        args += keys
        try:
            with cmd_pipe(self._base_args + args) as proc:
                return parse_keys(proc.stdout)
        except GPGError as e:
            if "error reading key" in str(e).lower():
                return []
            raise

    def encrypt(
        self,
        source: Optional[Union[bytes, io.FileIO, Callable]],
        recipients: List[str],
        output: Optional[Union[io.FileIO, Callable]],
        passphrase: Optional[str] = None,
        trust_model: TrustModel = TrustModel.pgp,
        sign: Optional[str] = None,
        compress_algo: Optional[CompressAlgo] = None,
        compress_level: Optional[int] = None,
    ):
        """Encrypt data

        Both source and output can be file like objects or callables
        processing streams.

        :param source: data to encrypt. Read directly from a file object
            or write the data to the stream in a callable.
        :param recipients: gpg key fingerprints
        :param output: encrypted data. Write directly to a file object
            or pass the stream to a callable.
        :param passphrase: passphrase of the private key to be used to sign
            the data.
        :param trust_model: trust model to be followed by gpg.
        :param sign: fingerprint of private key with which the
        data should be signed. If the parameter value is set to None the
        encrypted file is not signed.
        :param compress_algo: algorithm to use for data compression.
            If not provided the default algorithm defined by gpg is used.
        :param compress_level: data compression algorithm level.
            An integer value between 0 (no compression) and
            9 (maximum compression). If not provided the default level
            defined by gpg is used.
        :raises ValueError:
        """
        # Generate list of recipients for the encrypted file.
        args: Tuple[str, ...] = sum((("--recipient", rcp) for rcp in recipients), ())
        args += ("--no-tty", "--trust-model", trust_model.name, "--encrypt")
        if sign is not None:
            args += ("--sign", "--local-user", sign)
        if compress_algo is not None:
            args += ("--compress-algo", compress_algo.value)
        if compress_level is not None:
            if compress_level < 0 or compress_level > 9:
                raise ValueError(
                    "Invalid compression level. The value must "
                    "be an integer between 0 and 9."
                )
            args += ("--compress-level", str(compress_level))
        with cmd_pipe(
            self._base_args + self._pw_args(passphrase) + args,
            src=source,
            stdout=output,
            passphrase=passphrase,
        ):
            pass

    def decrypt(
        self,
        source: Union[io.FileIO, Callable],
        output: Union[io.FileIO, Callable],
        trust_model: TrustModel = TrustModel.pgp,
        passphrase: Optional[str] = None,
    ) -> List[str]:
        """Decrypt data and, if present, retrieve the signature appended to
        the encrypted file.

        :param source: data to decrypt. Read directly from a file object
            or write the data to the stream in a callable.
        :param output: decrypted data. Write directly to a file object
            or pass the stream to a callable.
        :param trust_model: trust model to be followed by gpg.
        :param passphrase: passphrase of the private key to be used to decrypt
            the data.
        :return: fingerprint or keyid of the signee's key.
        """

        # Note: the '--keyid-format', 'long' argument is only needed for
        # backwards compatibility with gpg versions < 2.1.0
        args = (
            "--batch",
            "--no-tty",
            "--status-fd",
            "2",
            "--trust-model",
            trust_model.name,
            "--decrypt",
        )
        maybe_stdout = {}
        if output is not None:
            maybe_stdout["stdout"] = output
        with cmd_pipe(
            self._base_args + self._pw_args(passphrase) + args,  # type: ignore
            **maybe_stdout,
            src=source,
            passphrase=passphrase,
        ) as proc:
            while True:
                if proc.poll() is not None:
                    break
            gpg_out = stderr_lookahead(proc)
            return capture_fpr(gpg_out)

    def gen_key(
        self,
        full_name: str,
        email: str,
        passphrase: str,
        key_length: int = 4096,
        key_type: str = "RSA",
    ):
        args = ("--batch", "--status-fd", "1", "--gen-key")
        with cmd_pipe(
            self._base_args + args,
            src=f"""
%echo Generating a basic OpenPGP key
Key-Type: {key_type}
Key-Length: {key_length}
Subkey-Type: {key_type}
Subkey-Length: {key_length}
Name-Real: {full_name}
Name-Email: {email}
Expire-Date: 0
Passphrase: {passphrase}
%commit
%echo done
""".encode(),
        ) as proc:
            l = next(
                line for line in proc.stdout if line.startswith(b"[GNUPG:] KEY_CREATED")
            )
            return l.rstrip(b"\r\n").split(b" ")[-1].decode("utf-8", "replace")

    def send_keys(
        self, *fingerprints: str, keyserver: str, url_opener=urllib.request.urlopen
    ):
        """Following https://tools.ietf.org/html/draft-shaw-openpgp-hkp-00#page-6, Section 4"""
        data = self.export(*fingerprints)
        post(
            build_host(keyserver) + "/pks/add",
            b"keytext=" + quote(data).encode(),
            url_opener=url_opener,
        )

    def recv_keys(self, *fingerprints: str, keyserver: str, **kwargs):
        for fingerprint in fingerprints:
            with download_key(
                keyserver=keyserver, fingerprint=fingerprint, **kwargs
            ) as key:
                self.import_file(key.read())

    def delete_keys(self, *fingerprints: str):
        cmd_devnull(
            self._base_args + ("--batch", "--yes", "--delete-keys") + fingerprints
        )

    def import_file(
        self, source: Union[bytes, IO[bytes]], *, passphrase: Optional[str] = None
    ):
        cmd_devnull(
            self._base_args + self._pw_args(passphrase) + ("--import", "--batch", "-"),
            src=source,
            passphrase=passphrase,
        )

    def export(self, *fingerprints: str, armor: bool = True) -> bytes:
        cmd = (
            self._base_args
            + ("--export",)
            + (("--armor",) if armor else ())
            + fingerprints
        )
        with cmd_pipe_stdout(cmd) as stream:
            return stream.read()

    def version(self):
        with cmd_pipe(self._base_args + ("--version",)) as proc:
            return tuple(
                map(
                    int,
                    proc.stdout.readline().rstrip(b"\n").split(b" ")[-1].split(b"."),
                )
            )

    def gen_revoke(self, fingerprint: str, passphrase) -> bytes:
        args = (
            self._base_args
            + self._pw_args(passphrase)
            + (
                "--no-tty",
                "--status-fd",
                "2",
                "--command-fd",
                "0",
                "--gen-revoke",
                fingerprint,
            )
        )
        with expect(args) as proc:
            try:
                proc.put(passphrase.encode() + b"\n")
                proc.expect(b"[GNUPG:] GET_BOOL gen_revoke.okay\n")
                proc.put(b"y\n")
                proc.expect(b"[GNUPG:] GET_LINE ask_revocation_reason.code\n")
                proc.put(b"0\n")
                proc.expect(b"[GNUPG:] GET_LINE ask_revocation_reason.text\n")
                proc.put(b"\n")
                proc.expect(b"[GNUPG:] GET_BOOL ask_revocation_reason.okay\n")
                proc.put(b"y\n")
            except ValueError as e:
                raise GPGError("Failed to generate a revocation certificate") from e
            return proc.stdout.read()

    def revoke(
        self, fingerprint: str, *, keyserver: Optional[str] = None, passphrase: str
    ):
        self.import_file(self.gen_revoke(fingerprint, passphrase))
        if keyserver is not None:
            self.send_keys(fingerprint, keyserver=keyserver)

    def detach_sig(
        self, src: Union[bytes, IO[bytes]], passphrase: str, user: Optional[str] = None
    ):
        return cmd_pipe_stdout(
            self._base_args
            + self._pw_args(passphrase)
            + (("--local-user", user) if user is not None else ())
            + ("--detach-sig", "-"),
            src=src,
            passphrase=passphrase,
        )

    def verify(self, src: Optional[Union[bytes, io.FileIO]], sig: bytes) -> str:
        """Verify and retrieve the pgp signature made to a file in detached
        signature mode - i.e. the signed file and the signature are in separate
        files.

        :param src: file that was signed.
        :param sig: detached signature file.
        :return: fingerprint or keyid of the signee's key.
        :raises GPGError:
        """
        f = tempfile.NamedTemporaryFile(delete=False)
        try:
            f.write(sig)
            f.seek(0)
            f.close()
            # Note: the '--keyid-format', 'long' argument is only needed for
            # backwards compatibility with gpg versions < 2.1.0
            with cmd_pipe(
                self._base_args + ("--status-fd", "2", "--verify", f.name, "-"), src=src
            ) as proc:
                gpg_out = stderr_lookahead(proc)
                fprs = capture_fpr(gpg_out)
                try:
                    (fpr,) = fprs
                except ValueError as e:
                    if not fprs:
                        raise GPGError(
                            "No fingerprint found in the output of gpg:\n"
                            + gpg_out.decode("utf-8", "replace")
                        ) from e
                    raise GPGError(
                        "Multiple fingerprints found in the output of gpg:\n"
                        + gpg_out.decode("utf-8", "replace")
                    ) from e
                return fpr
        finally:
            os.remove(f.name)


def search_keys(
    name: str, keyserver: str, url_opener=urllib.request.urlopen
) -> Iterator[KeyInfo]:
    try:
        queries = {normalize_fingerprint(name), name}
    except ValueError:
        queries = {name}
    for query in queries:
        try:
            with query_keyserver(
                keyserver=keyserver, query=query, op="index", url_opener=url_opener
            ) as response:
                yield from parse_search_keys(response)
        except HTTPError as e:
            if e.code != 404:
                raise e


def download_key(fingerprint: str, keyserver: str, **kwargs) -> ContextManager:
    return query_keyserver(
        keyserver=keyserver,
        query=normalize_fingerprint(fingerprint),
        op="get",
        exact="on",
        **kwargs,
    )


def query_keyserver(
    keyserver: str, query: str, op: str, url_opener=urllib.request.urlopen, **parameters
):
    url = build_host(keyserver)

    query_url = f"{url}/pks/lookup?search={quote(query)}&op={op}&options=mr" + "".join(
        f"?{key}={val}" for key, val in parameters.items()
    )
    return url_opener(query_url)


def split_host(url: str) -> Tuple[Optional[str], str, Optional[str]]:
    match = re.fullmatch(r"(https?://|hkp://)?([^:]+)(:[0-9]+)?", url)
    if not match:
        raise ValueError(f"Invalid URL: {url}")
    scheme, host, port = match.groups()
    return scheme, host, port


def build_host(url: str) -> str:
    scheme, host, port = split_host(url)
    if scheme not in {"http://", "https://"}:
        if port in {":80", ":8080", ":11371"}:
            scheme = "http://"
        else:
            scheme = "https://"
    if port is None:
        if scheme == "https://":
            port = ":443"
        else:
            port = ":80"
    return scheme + host + port


def extract_key_id(src: IO[bytes]) -> Iterator[str]:
    """Extract the public key_id from an encrypted message"""
    try:
        with preserve_pos(src):
            while True:
                fpr = extract_key_id_once(src)
                if fpr is None:
                    break
                yield fpr
    except ValueError as e:
        marker = b"-----BEGIN "
        src.seek(0)
        msg_start = src.read(len(marker))
        if msg_start == marker:
            raise GPGError("ASCII armored encrypted data is not supported") from e
        raise GPGError(f"Corrupted encrypted message: {e}") from e


def public_key_encrypted_session_key_packet(src: IO[bytes]) -> bytes:
    """type 1: Public-Key Encrypted Session Key Packet"""
    _version = src.read(1)
    return src.read(8)


def signature_packet(src: IO[bytes]) -> bytes:
    """Type 2: Signature Packet"""
    (version,) = src.read(1)
    if version == 3:
        (length,) = src.read(1)
        if length != 5:
            raise ValueError(f"Invalid length: {length}. Should be 5")
        _sigtype = src.read(1)
        _creationtime = src.read(4)
        return src.read(8)
    if version == 4:
        # fpr stars at octet 12
        _sigtype = src.read(1)
        _pubkeyalgorithm = src.read(1)
        _hashalgorithm = src.read(1)
        hashed_octet_count = int.from_bytes(src.read(2), "big")
        src.read(hashed_octet_count)
        unhashed_octet_count = int.from_bytes(src.read(2), "big")
        subpackets_end = src.tell() + unhashed_octet_count
        while src.tell() < subpackets_end:
            key_id = read_subpacket(src)
            if key_id:
                return key_id
        _signedhash = src.read(2)
        raise ValueError("No issuer subpacket in signature packet")
    raise ValueError(f"Invalid version: {version}. Only 3 and 4 are supported")


def read_subpacket(src: IO[bytes]) -> Optional[bytes]:
    l = read_new_packet_len(src)
    if l is None:
        return None
    (pkg_tpye,) = src.read(1)
    body = src.read(l - 1)
    if pkg_tpye == 16:  # type 16: Issuer
        return body
    return None


PACKET_HANDLERS = {1: public_key_encrypted_session_key_packet, 2: signature_packet}


def read_new_packet_len(src: IO[bytes]) -> Optional[int]:
    (l,) = src.read(1)
    if l < 192:  # 1 octet length
        return l
    if l <= 223:  # 2 octet length
        (l2,) = src.read(1)
        return ((l - 192) << 8) + l2 + 192
    # 224 <= l < 255: Partial Body Lengths (=> header has length 1)
    if l < 255:
        return None
    if l == 255:  # 5 octet length
        l2, l3, l4, l5 = src.read(4)
        return (l2 << 24) | (l3 << 16) | (l4 << 8) | l5
    raise ValueError(f"Invalid length: {l}")


def extract_key_id_from_sig(src: bytes) -> Optional[str]:
    """Extract key id from a signature packet. Either ascii armored or binary"""
    if src[0] == b"-"[0]:
        src = ascii_armored_to_binary(src)
    try:
        return extract_key_id_once(io.BytesIO(src))
    except ValueError as e:
        raise GPGError(
            "Found an invalid pgp signature while trying to extract a key id"
        ) from e


def extract_key_id_once(src: IO[bytes]) -> Optional[str]:
    """Reads one packet from src

    According to https://www.ietf.org/rfc/rfc4880.txt
    """
    try:
        (packet_hdr,) = src.read(1)
    except ValueError:
        return None
    # 1st bit must be 1
    if not packet_hdr & 0x80:
        raise ValueError("Not a valid packet tag")
    # 2nd bit is packet version (new / old)
    new_version = packet_hdr & 0x40
    packet_type = packet_hdr & 0b00111111
    if not new_version:
        packet_type >>= 2
    pkg_handler = PACKET_HANDLERS.get(packet_type)
    if pkg_handler is None:
        return None
    if not new_version:
        packet_length_type = packet_hdr & 0b11
        body_offs_by_len_type = {0: 2, 1: 3, 2: 5, 3: 1}
        try:
            packet_len_size = body_offs_by_len_type[packet_length_type] - 1
        except KeyError as e:
            raise ValueError(f"Invalid packet length type: {packet_length_type}") from e
        if packet_length_type == 3:  # The packet is of indeterminate length
            packet_len = None
        else:
            packet_len = int.from_bytes(src.read(packet_len_size), "big")
    else:
        packet_len = read_new_packet_len(src)
    if packet_len is not None and packet_len < 10:
        raise ValueError(f"Unexpected packet length of {packet_len}")
    body_start = src.tell()
    pub_key_id = pkg_handler(src)
    if packet_len:
        src.seek(body_start + packet_len)
    else:
        src.seek(0, 2)
    return pub_key_id.hex().upper()


def capture_fpr(b: bytes) -> List[str]:
    """Extract fingerprint from a standard output of gpg that looks like:
    [GNUPG:] NEWSIG
    gpg: Signature made Tue 10 Mar 2020 11:29:27 CET
    gpg:                using RSA key 3B3EB1983FAA63C1B05E072D066E042415D8FD9C
    [GNUPG:] KEY_CONSIDERED 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 0
    [GNUPG:] SIG_ID PaAuiNMUyUDRMsctXpV+7rX+0CI 2020-03-10 1583836167
    [GNUPG:] KEY_CONSIDERED 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 0
    [GNUPG:] GOODSIG 066E042415D8FD9C Chuck Norris <chucknorris@roundhouse.gov>
    gpg: Good signature from "Chuck Norris <chucknorris@roundhouse.gov>" [unknown]
    [GNUPG:] VALIDSIG 3B3EB1983FAA63C1B05E072D066E042415D8FD9C 2020-03-10 1583836167

    :param b: gpg stdout in binary form.
    :return: either a 40-char long fingerprint.
    """
    return [
        m.group(1).decode("utf-8", "replace")
        for m in re.finditer(rb"\[GNUPG:\] VALIDSIG ([0-9A-F]{40})", b)
    ]


@contextmanager
def preserve_pos(src: IO[bytes]):
    pos = src.tell()
    yield src
    src.seek(pos)


default_opener = urllib.request.build_opener(
    urllib.request.HTTPSHandler(debuglevel=0)
).open


def post(url, data, url_opener=default_opener):
    request = urllib.request.Request(
        url,
        data,
        method="POST",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    with url_opener(request) as response:
        response_text = response.read()
        return response_text


def ascii_armored_to_binary(src: bytes):
    header, _, *lines, footer = src.splitlines()
    # If the last line is empty, continue to the last not empty line:
    while not footer:
        footer = lines.pop()
    if header != b"-----BEGIN PGP SIGNATURE-----":
        raise ValueError(
            "Invalid ascii armored signature header: '" + header.decode() + "'"
        )
    if footer != b"-----END PGP SIGNATURE-----":
        raise ValueError(
            "Invalid ascii armored signature footer: '" + footer.decode() + "'"
        )
    return base64.decodebytes(b"".join(l for l in lines if l))



import unittest
from pathlib import Path
from unittest import mock
import io
from contextlib import contextmanager


class TestDocs(unittest.TestCase):
    @staticmethod
    def test_gen_key():
        """Read the python code snippets from README and exec"""
        read = False
        code = ""
        with open(Path(__file__).parent.parent / "README.md", encoding="utf-8") as doc:
            for line in doc:
                if line == "```python\n":
                    read = True
                    continue
                if line == "```\n":
                    read = False
                    continue
                if read:
                    code += line

        @contextmanager
        def mock_cmd(cmd, *_, **__):
            if "--version" in cmd:
                yield mock.Mock(
                    stdout=io.BytesIO(
                        b"""gpg (GnuPG) 2.2.19
libgcrypt 1.8.5
Copyright (C) 2019 F..."""
                    )
                )
            elif "--list-public-keys" in cmd:
                yield mock.Mock(
                    stdout=io.BytesIO(
                        b"""tru::1:1571400791:0:3:1:5
pub:u:2048:1:FDC66B7A5F6FFD4B:1550241679:::u:::scESC::::::23:1571230031:1 http\\x3a//keyserver.dcc.sib.swiss\\x3a11371:
fpr:::::::::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:
uid:u::::1550241679::B1E8D4565D26BED47D1EB8BB2B5733AEC1F154B5::Chuck Norris <chuck.norris@roundhouse.gov>::::::::::0:
sig:::1:FDC66B7A5F6FFD4B:1550241679::::Chuck Norris <chuck.norris@roundhouse.gov>:13x::CB5A3A5C2C4419BB09EFF879D6D229C51AB5D107:::8:
sig:::1:3A6500D5C1DE39AC:1571225229::::DCC Test (Test key for transferprotocol) <dcc@sib.swiss>:13x:::::2:
sig:?::1:2E3E7FE28FE412AE:1573661452::::[User ID not found]:13x::2866A142A16F54E06EB73CF92E3E7FE28FE412AE:::8:
sub:u:2048:1:D892C41917B20115:1550241679::::::e::::::23:
fpr:::::::::55C5314BB9EFD19AE7CC4774D892C41917B20115:"""
                    )
                )
            else:
                yield mock.Mock(stdout=io.BytesIO(b"[GNUPG:] KEY_CREATED aaaaaaaaaaaa"))

        with mock.patch("gpg_lite.cmd_pipe", mock_cmd), mock.patch(
            "gpg_lite.cmd_pipe_stdout", mock.Mock()
        ), mock.patch("gpg_lite.cmd_devnull", mock.Mock()), mock.patch(
            "builtins.print", mock.Mock()
        ):
            exec(code)  # pylint: disable=exec-used



#!/usr/bin/env python3

from datetime import datetime
from setuptools import setup, find_packages

# 0.0.0-dev.* version identifiers for development only (not public)
__version__ = "0.0.0.dev" + datetime.now().strftime("%Y%m%d")

setup(
    name="gpg-lite",
    version="0.8.0",
    license="LGPL3",
    description="python gpg bindings",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Jaroslaw Surkont, "
    "Gerhard Bräunlich, "
    "Robin Engler, "
    "Christian Ribeaud, "
    "François Martin",
    author_email="jaroslaw.surkont@unibas.ch, "
    "gerhard.braeunlich@id.ethz.ch, "
    "robin.engler@sib.swiss, "
    "christian.ribeaud@karakun.com, "
    "francois.martin@karakun.com",
    url="https://gitlab.com/biomedit/gpg-lite",
    python_requires=">=3.6",
    install_requires=["dataclasses ; python_version<'3.7'"],
    packages=find_packages(
        exclude=["test", "test.*", "integration_test", "integration_test.*"]
    ),
    package_data={"gpg_lite": ["py.typed"]},
    zip_safe=False,
    test_suite="test",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
)



import collections.abc
import sys
import warnings
from typing import Optional, Union, Tuple

import dataclasses


def deserialize(T: type):
    """Creates a deserializer for the type :T:. It handles dataclasses,
    sequences, typing.Optional and primitive types.

    :returns: A deserializer, converting a dict, list or primitive to :T:
    """

    def _deserialize(data):
        seq_type = sequence_type(T)
        if seq_type is not None:
            try:
                item_type = T.__args__[0]
            except AttributeError as e:
                raise ValueError(
                    f"Sequence of type {seq_type.__name__} without item type"
                ) from e
            return seq_type(map(deserialize(item_type), data))
        if dataclasses.is_dataclass(T):
            fields = dataclasses.fields(T)
            unexpected_keys = set(data.keys()) - set(f.name for f in fields)
            if unexpected_keys:
                warnings.warn(
                    f"{T.__name__}: Unexpected keys: " + ", ".join(unexpected_keys)
                )
            converted_data = {
                f.name: deserialize(f.type)(data.get(f.name, f.default)) for f in fields
            }
            return T(**converted_data)
        opt_type = optional_type(T)
        if opt_type is not None:
            if data is None:
                return None
            return opt_type(data)
        dct_type = dict_type(T)
        if dct_type is not None:
            key_type, val_type = dct_type
            return {
                deserialize(key_type)(key): deserialize(val_type)(val)
                for key, val in data.items()
            }

        return T(data)

    return _deserialize


def optional_type(T: type) -> Optional[type]:
    """For Optional types provided by typing, return the underlying python type,
    else None.
    E.g. returns int for typing.Optional[int]"""
    if getattr(T, "__origin__", None) is not Union:
        return None
    try:
        T1, T2 = T.__args__  # type: ignore
    except ValueError:
        return None
    if isinstance(None, T1):
        return T2
    if isinstance(None, T2):
        return T1
    return None


# Starting with python 3.7, the typing module has a new API
_origin_attr = "__extra__" if sys.version_info < (3, 7) else "__origin__"


def sequence_type(t: type) -> Optional[type]:
    """For sequence types provided by typing, return the underlying python type,
    else None.
    E.g. returns tuple for typing.Tuple[int]"""

    t_origin = getattr(t, _origin_attr, None)
    if isinstance(t_origin, type) and issubclass(t_origin, collections.abc.Sequence):
        return t_origin
    return None


def dict_type(t: type) -> Optional[Tuple[type, type]]:
    """For dicts provided by typing, return the underlying python types,
    else None.
    E.g. returns (int, str) for typing.Dict[int, str]"""

    t_origin = getattr(t, _origin_attr, None)
    if t_origin is dict:
        return t.__args__  # type: ignore
    return None
